/*
   A program which drops its privileges.  If it is setuid,
   it will drop its privileges to those of the user specified
   on the command line.
*/

#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <pwd.h>
#include <err.h>
#include <errno.h>
#include <stdlib.h>
#include <grp.h>
#include <string.h>

void print_sgroups(void)
{
    int nsg, i;
    gid_t *sg;

    nsg = getgroups(0, NULL); /* find out how many */
    if (nsg == -1)
        err(1, "getgroups(0,NULL) failed: ");
    if (nsg > 0) {
	if ((sg = (gid_t *)calloc(nsg, sizeof(gid_t))) == NULL) 
	    err(1, "calloc(%d,%d) failed: ", nsg, sizeof(gid_t));
	if ((nsg = getgroups(nsg, sg)) == -1) /* odd */
	    err(1, "getgroups(%d,%d) failed: ", nsg, sg);
	printf("    Supplemental groups: ");
	for (i=0; i<nsg-1; i++)
	    printf("%d, ", sg[i]);
	printf("%d\n", sg[nsg-1]);
    } else {
	printf("    Supplemental groups: none\n");
    }
}

void print_info(void)
{
    printf("    Real UID: %5d; Effective UID: %5d\n", getuid(), geteuid());
    printf("    Real GID: %5d; Effective GID: %5d\n", getgid(), getegid());
    print_sgroups();
}

int
main(int argc, char *argv[])
{
    struct passwd *newid;
    int rc1, rc2;

    if (argc != 2) {
        fprintf(stderr, "Usage: %s user\n", argv[0]);
	exit(1);
    }

    printf("Before change:\n");
    print_info();

    errno = 0;
    if ((newid = getpwnam(argv[1])) == NULL) {
        fprintf(stderr, "User '%s' lookup failed: %s\n", argv[0],
	        strerror(errno));
	if (errno == 0) fprintf(stderr, "User not found\n");
	exit(1);
    }
    endpwent();

    if ((rc1 = setgroups(0,0)) < 0)
        err(1, "setgroups(0,0) failed: ");
    if ((rc1 = setgid(newid->pw_gid)) < 0 || (rc2 = setuid(newid->pw_uid)) < 0) {
        fprintf(stderr, "Setting new uid (%d) and gid (%d) failed: %s\n", 
	        newid->pw_uid, newid->pw_gid, strerror(errno));
	fprintf(stderr, "Return codes: gid: %d; uid: %d\n", rc1, rc2);
	exit(1);
    }

    printf("After change:\n");
    print_info();
}
